#version 100

uniform sampler2D TextureYUV;
uniform sampler2D TextureMask;
uniform sampler2D TextureBlurred;
uniform lowp float Alpha;
uniform lowp float OutsideCropAlpha;
#ifndef __APPLE__
uniform mediump vec4 CropRect;
uniform mediump vec4 ThumbnailOffsetScale;
#else
uniform lowp vec4 CropRect;
uniform lowp vec4 ThumbnailOffsetScale;
#endif
uniform lowp float Exposure;
uniform lowp vec4 Parameters; // [Ty, Ysh, Yb, Yw]

#ifndef __APPLE__
varying mediump vec2 TexCoordInterpolated;
varying mediump vec2 PositionInterpolated;
#else
varying lowp vec2 TexCoordInterpolated;
varying lowp vec2 PositionInterpolated;
#endif

//
// According to: http://www.equasys.de/colorconversion.html (Note: column major layout)
//
const lowp mat3 yuvFullRangeToRgb = mat3(1.0, 1.0, 1.0,
                                         0.0, -0.343, 1.765,
                                         1.4, -0.711, 0.0);

void gamutCorrect(inout lowp vec3 yuv) {
    // Denominators of the red, green, and blue out-of-gamut equations
    lowp vec2 uvSigned = yuv.yz - vec2(0.5);
    
    // Note: To work around a strange issue on iPhone5C/iOS8 we're doing matrix mulitplication manually here instead of relying on mat3*vec3 built-in functionality
    lowp vec3 denom = vec3(1.4 * uvSigned.y,
                           -0.343 * uvSigned.x - 0.711 * uvSigned.y,
                           1.765 * uvSigned.x);

    // Avoid division by zero
    lowp vec3 divByZero = (vec3(1.0) - abs(sign(denom))) * 0.001;
    denom = denom + divByZero;
    
    // Without flipping the UV direction, only one kind of out-of-gamut can happen
    // Either over-saturation or under-saturation will produce a non-negative lambda
    lowp vec3 lambda_vect = max(vec3(1.0 - yuv.x) / denom, vec3(-yuv.x) / denom);

    // UVs will never be scaled outward toward the gamut edge
    lambda_vect = min(lambda_vect, vec3(1.0));

    // The smallest of the three is the maximum legal scale for the UV
    lowp float lambda = min(min(lambda_vect.x, lambda_vect.y), lambda_vect.z);

    // Apply adjusted UV to input
    yuv.yz = lambda * uvSigned + vec2(0.5);
}

//
// Remove background
//
lowp vec3 removeBackground(lowp vec3 yuv, lowp float mask) {

    lowp vec3 yuvBiased = yuv;

    // Y Bias
    lowp float Ty  = Parameters.x;
    lowp float B = 1.0 - Ty;

    yuvBiased.x = clamp(yuv.x + B, 0.0, 1.0);

    gamutCorrect(yuvBiased);
    
    // Blend to get the image with it's background removed
    return mix(yuvBiased, yuv, vec3(mask)); // Note: 0=>White background, 1=>Foreground
}

//
// Sharpen background/foreground transition
//
lowp float sharpenEdges(lowp float y, lowp float blurred, lowp float mask) {
    
    // Unsharp mask
    lowp float ySharp = clamp(1.5 * y - 0.5 * blurred, 0.0, 1.0);
    
    // Blend the sharpened Y value in on the edges
    lowp float edgeStrength = abs(0.5 - mask) * 2.0;
    return mix(ySharp, y, edgeStrength); // Note: 0=>Edge, 1=>Non-edge
}

//
// Improve saturation
//
lowp vec3 saturationTransform(lowp vec3 yuv, lowp float strength) {
    const lowp float uv_deadzone = 0.012; // Obtained empirically
    const lowp float uv_rampend = 0.036;  //
    const lowp float end_sigma = 6.0;     //

    const lowp float is2 = 0.70710678118; //inversesqrt(2.0);

    // Scale U, V to [1, 1]
    yuv.yz = 2.0 * yuv.yz - vec2(1.0);

    // UV "saturation" as a magnitude ranging from 0 to 1
    lowp float uv = length(yuv.yz) * is2;

    // This coordinate gives location along a sigmoid curve, which is how saturation will be assigned
    lowp float t = (uv - uv_deadzone) * end_sigma / (1.0 - uv_deadzone);
    t = max(t, 0.0);

    // The replacement function will be more extreme as Exposure is increased
    lowp float sigmoid = (1.0 - uv_deadzone) * (2.0 / (1.0 + exp(-t)) - 1.0) + uv_deadzone;
    lowp float f = mix(uv, sigmoid, strength * Exposure);

    // Values below uv_deadzone are not remapped. Between rampend and deadzone, the sigmoid is gradually applied.
    lowp float a = (uv - uv_deadzone) / (uv_rampend - uv_deadzone);
    a = clamp(a, 0.0, 1.0);

    // UV magnitude remap
    lowp float uv_new = mix(uv, f, a);
    yuv.yz = yuv.yz * (uv_new / (uv + 0.0001));

    // Scale U, V back to [0, 1] and ensure values stay in gamut
    yuv.yz = 0.5 * (yuv.yz + vec2(1.0));
    gamutCorrect(yuv);

    return yuv;
}

//
// Apply local exposure curve
//
lowp float applyCurve(lowp float y, lowp float blurred) {
    
    lowp float Ty  = Parameters.x;
    lowp float Ysh = Parameters.y;
    lowp float Yb  = Parameters.z;
    lowp float Yw  = Parameters.w;
    
    // Apply sigmoid curve
    lowp float t = 6.0 * (y - Yb) / (Ysh - Yb);
    lowp float ySigmoid = clamp(Ysh * (2.0 / (1.0 + exp(-t)) - 1.0), 0.0, 1.0);

    // Apply linear curve (black and white point)
    lowp float yLinear = clamp((y - Yb) / (Yw - Yb), 0.0, 1.0);

    // Generate combined curve output
    lowp float blend = clamp((Ysh - blurred) / (Ysh - Yb), 0.0, 1.0);
    lowp float yWithCurve = mix(yLinear, ySigmoid, blend);

    // Blend with original, according to 'Exposure' (0=>Original, 1=>Curve)
    return mix(y, yWithCurve, Exposure);
}

lowp float applyCurveWObj(lowp float y, lowp float blurred) {
    lowp float Yw  = Parameters.x;
    lowp float Ysh = Exposure * Parameters.y + 0.0001;
    lowp float Yb  = Exposure * Parameters.z;

    // Apply sigmoid curve
    lowp float t = 6.0 * (y - Yb) / (Ysh - Yb);
    lowp float ySigmoid = clamp(Ysh * (2.0 / (1.0 + exp(-t)) - 1.0), 0.0, 1.0);

    // Apply gamma curve (black and white point with exponent)
    lowp float yGamma = clamp((y - Yb) / (Yw - Yb), 0.0, 1.0);
    yGamma = pow(yGamma, 1.0 - 0.4 * Exposure);

    // Generate combined curve output
    lowp float blend = clamp((Ysh - blurred) / (Ysh - Yb), 0.0, 1.0);
    return mix(yGamma, ySigmoid, blend);
}

//
// Main entry point
//
void main(void) {
    
    // Get input YUV value
    lowp vec3 yuv0 = texture2D(TextureYUV, TexCoordInterpolated).rgb;
    
    // Get background mask and blurred YUV values
    lowp vec2 ThumbnailTexCoord = TexCoordInterpolated * ThumbnailOffsetScale.zw + ThumbnailOffsetScale.xy;
    lowp vec4 mask = texture2D(TextureMask, ThumbnailTexCoord);
    lowp vec4 blurred = texture2D(TextureBlurred, ThumbnailTexCoord);

    lowp vec3 yuvResult;
    
#ifdef WHITE_OBJECT
    
    // Step 1/1: Apply curve (Y channel only)
    lowp float y1 = applyCurveWObj(yuv0.r, blurred.r);
    
    // Combine result
    yuvResult = vec3(y1, yuv0.gb);
    
    // Saturation transform
    yuvResult = saturationTransform(yuvResult, 0.6);
    
#else
    
    // Step 1/3: Remove background
    lowp vec3 yuv1 = removeBackground(yuv0, mask.r);
    
    // Step 2/3: Sharpen edges (Y channel only)
    lowp float y2 = sharpenEdges(yuv1.r, blurred.r, mask.r);
    
    // Step 3/3: Apply curve (Y channel only)
    lowp float y3 = applyCurve(y2, blurred.r);
    
    // Combine result
    yuvResult = vec3(y3, yuv1.gb);
    
    // Saturation transform
    yuvResult = saturationTransform(yuvResult, 0.55);
    
#endif

#ifdef RETURN_YUV
    
    // Assign YUV output
    gl_FragColor = vec4(yuvResult, 1.0);
    
#else
    
    // Convert blended YUV sample to RGB
    yuvResult.y -= 0.5;
    yuvResult.z -= 0.5;
    lowp vec3 rgb = yuvFullRangeToRgb * yuvResult;
    
    // Get alpha value for crop visualization
    bool inside = (PositionInterpolated.x > CropRect.x && PositionInterpolated.x < CropRect.y &&
                    PositionInterpolated.y > CropRect.z && PositionInterpolated.y < CropRect.w);
    lowp float cropAlpha = float(inside) + OutsideCropAlpha * float(!inside);
    
    // Assign RGBA output
    gl_FragColor = vec4(rgb, Alpha * cropAlpha);
    
#endif
}
